@@ -3,8 +3,7 @@
 /*               This file is part of the program and library                */
 /*    PaPILO --- Parallel Presolve for Integer and Linear Optimization       */
 /*                                                                           */
-/* Copyright (C) 2020-2023 Konrad-Zuse-Zentrum                               */
-/*                     fuer Informationstechnik Berlin                       */
+/* Copyright (C) 2020-2024 Zuse Institute Berlin (ZIB)                       */
 /*                                                                           */
 /* This program is free software: you can redistribute it and/or modify      */
 /* it under the terms of the GNU Lesser General Public License as published  */
@@ -26,8 +25,9 @@
 
 #include "papilo/core/Presolve.hpp"
 #include "papilo/core/postsolve/Postsolve.hpp"
-#include "papilo/io/MpsParser.hpp"
+#include "papilo/io/Parser.hpp"
 #include "papilo/io/MpsWriter.hpp"
+#include "papilo/io/OpbWriter.hpp"
 #include "papilo/io/SolParser.hpp"
 #include "papilo/io/SolWriter.hpp"
 #include "papilo/misc/NumericalStatistics.hpp"
@@ -61,7 +61,8 @@ ResultStatus
 presolve_and_solve(
     const OptionsInfo& opts,
     std::unique_ptr<SolverFactory<REAL>> lpSolverFactory = nullptr,
-    std::unique_ptr<SolverFactory<REAL>> mipSolverFactory = nullptr )
+    std::unique_ptr<SolverFactory<REAL>> mipSolverFactory = nullptr,
+    std::unique_ptr<SolverFactory<REAL>> satSolverFactory = nullptr )
 {
    try
    {
@@ -71,7 +72,7 @@ presolve_and_solve(
 
       {
          Timer t( readtime );
-         prob = MpsParser<REAL>::loadProblem( opts.instance_file );
+         prob = Parser<REAL>::loadProblem( opts.instance_file );
       }
 
       // Check whether reading was successful or not
@@ -99,6 +100,8 @@ presolve_and_solve(
             lpSolverFactory->add_parameters(paramSet);
          if(mipSolverFactory != nullptr)
             mipSolverFactory->add_parameters(paramSet);
+         if(satSolverFactory != nullptr)
+            satSolverFactory->add_parameters(paramSet);
 
          if( !opts.param_settings_file.empty() && !opts.print_params )
          {
@@ -205,11 +208,12 @@ presolve_and_solve(
 
       presolve.setLPSolverFactory( std::move( lpSolverFactory ) );
       presolve.setMIPSolverFactory( std::move( mipSolverFactory ) );
+      presolve.setSATSolverFactory( std::move( satSolverFactory ) );
 
       presolve.getPresolveOptions().tlim =
           std::min( opts.tlim, presolve.getPresolveOptions().tlim );
 
-      bool store_dual_postsolve = false;
+      bool store_dual = false;
       std::unique_ptr<SolverInterface<REAL>> solver;
       if(opts.command == Command::kSolve)
       {
@@ -218,7 +222,18 @@ presolve_and_solve(
          {
             solver = presolve.getLPSolverFactory()->newSolver(
                 presolve.getVerbosityLevel() );
-            store_dual_postsolve = solver->is_dual_solution_available();
+            store_dual = solver->is_dual_solution_available();
+         }
+         else if( presolve.getSATSolverFactory() &&
+                  problem.test_problem_type(ProblemFlag::kBinary) )
+         {
+            if( !presolve.getPresolveOptions().verification_with_VeriPB )
+            {
+               fmt::print( "please activate VeriPB to provide row scaling for SAT Solvers\n" );
+               return ResultStatus::kError;
+            }
+            solver = presolve.getSATSolverFactory()->newSolver(
+                presolve.getVerbosityLevel() );
          }
          else if( presolve.getMIPSolverFactory() )
             solver = presolve.getMIPSolverFactory()->newSolver(
@@ -230,7 +245,7 @@ presolve_and_solve(
          }
       }
 
-      auto result = presolve.apply( problem, store_dual_postsolve );
+      auto result = presolve.apply( problem, store_dual );
 
       if( !opts.optimal_solution_file.empty() )
       {
@@ -275,9 +290,26 @@ presolve_and_solve(
       {
          Timer t( writetime );
 
-         MpsWriter<REAL>::writeProb( opts.reduced_problem_file, problem,
-                                     result.postsolve.origrow_mapping,
-                                     result.postsolve.origcol_mapping );
+         if(boost::algorithm::ends_with( opts.reduced_problem_file, ".opb" )
+             || boost::algorithm::ends_with( opts.reduced_problem_file, ".opb.bz2" )
+             || boost::algorithm::ends_with( opts.reduced_problem_file, ".opb.gz" ))
+         {
+            Num<REAL> num{};
+            num.setFeasTol( REAL{ presolve.getPresolveOptions().feastol } );
+            num.setEpsilon( REAL{ presolve.getPresolveOptions().epsilon } );
+            num.setHugeVal( REAL{ presolve.getPresolveOptions().hugeval } );
+            bool success = OpbWriter<REAL>::writeProb( opts.reduced_problem_file, problem,
+                                        result.postsolve.origcol_mapping, presolve.getRowScalingFactors(), num );
+            //TODO: change name
+            if(!success)
+               MpsWriter<REAL>::writeProb( opts.reduced_problem_file, problem,
+                                           result.postsolve.origrow_mapping,
+                                           result.postsolve.origcol_mapping );
+         }
+         else
+            MpsWriter<REAL>::writeProb( opts.reduced_problem_file, problem,
+                                        result.postsolve.origrow_mapping,
+                                        result.postsolve.origcol_mapping );
 
          fmt::print( "reduced problem written to {} in {:.3f} seconds\n\n",
                      opts.reduced_problem_file, t.getTime() );
@@ -286,6 +318,7 @@ presolve_and_solve(
       if( !opts.postsolve_archive_file.empty() )
       {
 
+#ifdef PAPILO_SERIALIZATION_AVAILABLE
          Timer t( writetime );
          std::ofstream ofs( opts.postsolve_archive_file,
                             std::ios_base::binary );
@@ -295,18 +328,24 @@ presolve_and_solve(
          oa << result.postsolve;
          fmt::print( "postsolve archive written to {} in {:.3f} seconds\n\n",
                      opts.postsolve_archive_file, t.getTime() );
+#else
+         fmt::print( "writing to potsolve archive is not possible without Boost Serialization!\n");
+#endif
       }
 
-      if( opts.command == Command::kPresolve )
+      if( opts.command == Command::kPresolve || problem.getNCols() == 0 )
          return ResultStatus::kOk;
 
       double solvetime = 0;
       {
          Timer t( solvetime );
 
+         if(presolve.getPresolveOptions().verification_with_VeriPB)
+            solver->setRowScalingFactor(presolve.getRowScalingFactors());
          solver->setUp( problem, result.postsolve.origrow_mapping,
                         result.postsolve.origcol_mapping );
 
+
          if( opts.tlim != std::numeric_limits<double>::max() )
          {
             double tlim =
@@ -329,39 +368,50 @@ presolve_and_solve(
          Solution<REAL> solution;
          solution.type = SolutionType::kPrimal;
 
-         if( result.postsolve.getOriginalProblem().getNumIntegralCols() == 0 &&
-             store_dual_postsolve )
+         if( result.postsolve.getOriginalProblem().getNumIntegralCols() == 0 && store_dual )
             solution.type = SolutionType::kPrimalDual;
 
          if( ( status == SolverStatus::kOptimal ||
                status == SolverStatus::kInterrupted ) &&
-             solver->getSolution( solution ) )
+             solver->getSolution( solution, result.postsolve ) )
             postsolve( result.postsolve, solution, opts.objective_reference,
                        opts.orig_solution_file, opts.orig_dual_solution_file,
                        opts.orig_reduced_costs_file, opts.orig_basis_file );
-
-         if( status == SolverStatus::kInfeasible )
+         solvetime = t.getTime();
+         double time =
+             presolve.getStatistics().presolvetime + solvetime + writetime;
+         fmt::print("Solving time {:.3f} seconds\n", time);
+         switch( status )
+         {
+      case SolverStatus::kInfeasible:
             fmt::print(
                 "\nsolving detected infeasible problem after {:.3f} seconds\n",
-                presolve.getStatistics().presolvetime + solvetime + writetime );
-         else if( status == SolverStatus::kUnbounded )
+                time );
+            break;
+         case SolverStatus::kUnbounded:
             fmt::print(
                 "\nsolving detected unbounded problem after {:.3f} seconds\n",
-                presolve.getStatistics().presolvetime + solvetime + writetime );
-         else if( status == SolverStatus::kUnbndOrInfeas )
+                time );
+            break;
+         case SolverStatus::kUnbndOrInfeas:
             fmt::print(
                 "\nsolving detected unbounded or infeasible problem after "
                 "{:.3f} seconds\n",
-                presolve.getStatistics().presolvetime + solvetime + writetime );
-         else
-            fmt::print( "\nsolving finished after {:.3f} seconds\n",
-                        presolve.getStatistics().presolvetime + solvetime +
-                            writetime );
+                time );
+            break;
+         case SolverStatus::kInterrupted:
+         case SolverStatus::kError:
+            fmt::print( "\nsolving interrupted after {:.3f} seconds\n", time );
+            break;
+         default:
+            fmt::print( "\nsolving finished after {:.3f} seconds\n", time );
+            break;
+         }
       }
    }
    catch( std::bad_alloc& ex )
    {
-      fmt::print( "Memory out exception occured! Please assign more memory\n" );
+      fmt::print( "Memory out exception occurred! Please assign more memory\n" );
       return ResultStatus::kError;
    }
 
@@ -445,6 +495,14 @@ postsolve( PostsolveStorage<REAL>& postsolveStorage,
                   primal_solution_output, sec3 );
    }
 
+   if( ( !reduced_solution_output.empty() || !dual_solution_output.empty() ||
+         !basis_output.empty() ) &&
+       original_sol.type == SolutionType::kPrimal )
+   {
+      fmt::print(
+          "dual solution does not exist and therefore can not be saved.\n" );
+   }
+
    if( !dual_solution_output.empty() &&
        original_sol.type == SolutionType::kPrimalDual )
    {
@@ -543,6 +601,7 @@ template <typename REAL>
 void
 postsolve( const OptionsInfo& opts )
 {
+#ifdef PAPILO_SERIALIZATION_AVAILABLE
    PostsolveStorage<REAL> ps;
    std::ifstream inArchiveFile( opts.postsolve_archive_file,
                                 std::ios_base::binary );
@@ -551,15 +610,73 @@ postsolve( const OptionsInfo& opts )
    inArchiveFile.close();
 
    SolParser<REAL> parser;
-   Solution<REAL> reduced_solution;
+   Vec<REAL> primal_solution;
    bool success = parser.read( opts.reduced_solution_file, ps.origcol_mapping,
                                ps.getOriginalProblem().getVariableNames(),
-                               reduced_solution );
+                               primal_solution );
+   Solution<REAL> reduced_solution{primal_solution};
    if( success )
+   {
+      if( !opts.reduced_dual_solution_file.empty() !=
+          !opts.reduced_reduced_costs_solution_file.empty() )
+      {
+         fmt::print( "Dual solution and reduced costs must be provided. Only "
+                     "original primal solution reconstructed.\n" );
+      }
+      else if( !opts.reduced_dual_solution_file.empty() &&
+          !opts.reduced_reduced_costs_solution_file.empty() )
+      {
+         if(ps.postsolveType == PostsolveType::kFull)
+         {
+            bool dual_success = parser.read(
+                opts.reduced_dual_solution_file, ps.origrow_mapping,
+                ps.getOriginalProblem().getConstraintNames(),
+                reduced_solution.dual );
+
+            if( !dual_success )
+               return;
+
+            bool reduced_costs_success = parser.read(
+                opts.reduced_reduced_costs_solution_file, ps.origcol_mapping,
+                ps.getOriginalProblem().getVariableNames(),
+                reduced_solution.reducedCosts );
+            if( !reduced_costs_success )
+               return;
+            reduced_solution.type = SolutionType::kPrimalDual;
+
+            if( !opts.reduced_basis_solution_file.empty() && ps.presolveOptions.calculate_basis_for_dual )
+            {
+                  reduced_solution.basisAvailabe = true;
+                  parser.read_basis( opts.reduced_basis_solution_file, ps,
+                                     reduced_solution.varBasisStatus,
+                                     reduced_solution.rowBasisStatus );
+            }
+            else
+            {
+               if(!opts.reduced_basis_solution_file.empty())
+                  fmt::print("Postsolve storage not suitable for basis calculation.\n");
+               reduced_solution.basisAvailabe = false;
+               reduced_solution.varBasisStatus =
+                   Vec<VarBasisStatus>{ reduced_solution.primal.size() };
+               reduced_solution.rowBasisStatus =
+                   Vec<VarBasisStatus>( reduced_solution.dual.size() );
+            }
+         }
+         else
+         {
+            fmt::print(
+                "Postsolve does not contain information about dual solution. "
+                "Only original primal solution reconstructed.\n" );
+         }
+      }
       postsolve( ps, reduced_solution, opts.objective_reference,
                  opts.orig_solution_file, opts.orig_dual_solution_file,
                  opts.orig_reduced_costs_file, opts.orig_basis_file );
-}
+
+
+   }
+#endif
+   }
 
 } // namespace papilo
 
